
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static void	Error(char *msg)
{
	(void)fputs(msg,stderr); (void)fputc('\n',stderr); exit(1);
}

static FILE	*OpenRead(char *path)
{
	FILE	*file=(*path=='-')?stdin:fopen(path,"r");

	if (file==NULL) {
	    (void)fputs(path,stderr); Error(" : file not found.");
	}
	return file;
}

#define MALLOC(type,size)	(type *)malloc(sizeof(type)*(size))

static char	*mae="memory allocation error.";

#define UL	unsigned long

static UL	Ne;

typedef struct Edge {
		UL		name,sv,ev,rf,lf;
		struct Edge	*sce,*scce,*ece,*ecce,*next;
	} Edge;

#ifndef	HASH_SIZE
#define HASH_SIZE	1024
#endif

#ifndef	STATIC_HASH
static Edge	**hash;
#else
static Edge	*hash[HASH_SIZE*HASH_SIZE];
#endif

static void	InitHash()
{
	size_t	i;

#ifndef	STATIC_HASH
	if ((hash=MALLOC(Edge *,HASH_SIZE*HASH_SIZE))==NULL) Error(mae);
#endif
	Ne=0; for (i=0; i<HASH_SIZE*HASH_SIZE; i++) hash[i]=NULL;
}

#define EMPTY	(UL)(~0)

static Edge	*SearchEdge(UL tail,UL head,UL f)
{
	Edge	**last,*e;
	UL	sv,ev,rf,lf;

	if (tail<head) {
	    last=hash+(tail%HASH_SIZE)*HASH_SIZE+head%HASH_SIZE;
	    while ((e=(*last))!=NULL) {
		if (e->sv==tail && e->ev==head && e->lf==EMPTY) {
		    e->lf=f; return e;
		}
		last=(&(e->next));
	    }
	    sv=tail; ev=head; rf=EMPTY; lf=f;
	}
	else {
	    last=hash+(head%HASH_SIZE)*HASH_SIZE+tail%HASH_SIZE;
	    while ((e=(*last))!=NULL) {
		if (e->sv==head && e->ev==tail && e->rf==EMPTY) {
		    e->rf=f; return e;
		}
		last=(&(e->next));
	    }
	    sv=head; ev=tail; rf=f; lf=EMPTY;
	}
	if (Ne==EMPTY) Error("too many edges.");

	if ((*last=e=MALLOC(Edge,1))==NULL) Error(mae);

	e->name=Ne++; e->sv=sv; e->ev=ev; e->rf=rf; e->lf=lf;
	e->sce=e->scce=e->ece=e->ecce=e->next=NULL; return e;
}

static void	SetEdge(UL tail,UL head,Edge *e,Edge *succ,Edge *prev)
{
	if (tail<head) {
	    e->ece=succ; e->scce=prev;
	}
	else {
	    e->sce=succ; e->ecce=prev;
	}
}

static Edge	**TermHash()
{
	Edge	**edge,*e;
	size_t	i;

	if ((edge=MALLOC(Edge *,Ne))==NULL) Error(mae);

	for (i=0; i<HASH_SIZE*HASH_SIZE; i++)
	    for (e=hash[i]; e!=NULL; e=e->next) edge[e->name]=e;

#ifndef	STATIC_HASH
	free(hash);
#endif
	return edge;
}

static FILE	*OpenWrite(char *path,UL N)
{
	FILE	*file=(*path=='-')?stdout:fopen(path,"w");

	if (file==NULL) {
	    (void)fputs(path,stderr); Error(" : file not open.");
	}
	(void)fprintf(file,"%lu\n",N); return file;
}

static void	PrintVFE(FILE *file,UL vfe,int t_n)
{
	if (vfe==EMPTY)
	    (void)fprintf(file,"-%c",t_n);
	else
	    (void)fprintf(file,"%lu%c",vfe,t_n);
}

static void	PrintEdge(FILE *file,Edge *e,int t_n)
{
	if (e==NULL)
	    (void)fprintf(file,"-%c",t_n);
	else
	    (void)fprintf(file,"%lu%c",e->name,t_n);
}

#define LEN	2048

int	main(int argc,char **argv)
{
	FILE	*file;
	char	str[LEN];
	UL	Nv,Nf,*VE,*FE,i,v1,v2,v3;
	Edge	*e1,*e2,*e3,**edge,*e;

	if (argc!=2 && argc!=3 && argc!=5 && argc!=6)
	    Error("usage : vf2wes VF|(V F) {VE FE WES}");

	file=OpenRead(argv[1]);
	if (argc==2 || argc==5) {
	    if (fgets(str,LEN,file)==NULL || str[strlen(str)-1]!='\n' ||
		sscanf(str,"%lu %lu",&Nv,&Nf)!=2 || Nv==0 || Nf==0)
		Error("bad number of vertices and/or faces.");
	}
	else
	    if (fgets(str,LEN,file)==NULL || str[strlen(str)-1]!='\n' ||
		sscanf(str,"%lu",&Nv)!=1 || Nv==0)
		Error("bad number of vertices.");

	if ((VE=MALLOC(UL,Nv))==NULL) Error(mae);

	for (i=0; i<Nv; i++) {
	    if (fgets(str,LEN,file)==NULL || str[strlen(str)-1]!='\n')
		Error("vertex read error.");

	    VE[i]=EMPTY;
	}
	if (argc!=2 && argc!=5) {
	    if (file!=stdin) (void)fclose(file);

	    file=OpenRead(argv[2]);
	    if (fgets(str,LEN,file)==NULL || str[strlen(str)-1]!='\n' ||
		sscanf(str,"%lu",&Nf)!=1 || Nf==0)
		Error("bad number of faces.");
	}
	if ((FE=MALLOC(UL,Nf))==NULL) Error(mae);

	InitHash();
	for (i=0; i<Nf; i++) {
	    if (fgets(str,LEN,file)==NULL || str[strlen(str)-1]!='\n' ||
		sscanf(str,"%lu %lu %lu",&v1,&v2,&v3)!=3 ||
		v1>=Nv || v2>=Nv || v3>=Nv) Error("face read error.");

	    e1=SearchEdge(v1,v2,i);
	    e2=SearchEdge(v2,v3,i);
	    e3=SearchEdge(v3,v1,i); SetEdge(v1,v2,e1,e2,e3);
				    SetEdge(v2,v3,e2,e3,e1);
				    SetEdge(v3,v1,e3,e1,e2); FE[i]=EMPTY;
	}
	edge=TermHash(); if (file!=stdin) (void)fclose(file);

	for (i=0; i<Ne; i++) {
	    e=edge[i];
	    if (VE[e->sv]==EMPTY) VE[e->sv]=e->name;
	    if (VE[e->ev]==EMPTY) VE[e->ev]=e->name;
	    if (e->rf!=EMPTY && FE[e->rf]==EMPTY) FE[e->rf]=e->name;
	    if (e->lf!=EMPTY && FE[e->lf]==EMPTY) FE[e->lf]=e->name;
	}
	if (argc==2 || argc==3) {
	    file=stdout; (void)printf("%lu\t%lu\t%lu\n",Nv,Nf,Ne);
	}
	else
	    file=OpenWrite(argv[argc-3],Nv);

	for (i=0; i<Nv; i++) PrintVFE(file,VE[i],'\n');

	if (argc!=2 && argc!=3) {
	    if (file!=stdout) (void)fclose(file);

	    file=OpenWrite(argv[argc-2],Nf);
	}
	for (i=0; i<Nf; i++) PrintVFE(file,FE[i],'\n');

	if (argc!=2 && argc!=3) {
	    if (file!=stdout) (void)fclose(file);

	    file=OpenWrite(argv[argc-1],Ne);
	}
	for (i=0; i<Ne; i++) {
	    e=edge[i];
	    PrintVFE(file,e->sv,'\t'); PrintVFE(file,e->ev,'\t');
	    PrintVFE(file,e->rf,'\t'); PrintVFE(file,e->lf,'\t');
	    PrintEdge(file,e->sce,'\t'); PrintEdge(file,e->scce,'\t');
	    PrintEdge(file,e->ece,'\t'); PrintEdge(file,e->ecce,'\n');
	}
	if (file!=stdout) (void)fclose(file);

	for (i=0; i<Ne; i++) free(edge[i]); free(edge); free(FE); free(VE);

	return 0;
}
